/* ircd_auth.c - NextIRCd
 * Copyright (C) 2012-2013  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#include <string.h>

#include "ircd_auth.h"
#include "ircd_sock.h"
#include "struct_auth.h"
#include "struct_connection.h"
#include "ircd_client.h"
#include "IOHandler/IOSockets.h"
#include "IOHandler/IODNSLookup.h"
#include "version.h"

static void auth_start_dnsreverse(struct Auth *auth);
static IODNS_CALLBACK(auth_dns_callback);
static void auth_free(struct Auth *auth);

static struct Auth *authlist_first = NULL;
static struct Auth *authlist_last = NULL;

struct Auth *auth_new(struct Connection *conn) {	
	struct Auth *auth = calloc(1, sizeof(*auth));
	socket_printf(conn, "NOTICE AUTH :*** NextIRCd v%d.%d (%s)", VERSION_NUMBER, patchlevel, revision);
	
	auth->conn = conn;
	time(&auth->startup_time);
	
	auth->prev = authlist_last;
	auth->next = NULL;
	if(!authlist_last)
		authlist_first = auth;
	authlist_last = auth;
	
	auth_start_dnsreverse(auth);
	
	return auth;
}

static void auth_start_dnsreverse(struct Auth *auth) {
	socket_printf(auth->conn, "NOTICE AUTH :*** Looking up your hostname");
	
	struct IODNSAddress *sockaddr;
	sockaddr = iosocket_get_remote_addr(auth->conn->socket);
	if(sockaddr) {
		auth->dnslookup = iodns_getnameinfo(sockaddr->address, sockaddr->addresslen, auth_dns_callback, auth);
	} else {
		// critical error!
	}
}

static IODNS_CALLBACK(auth_dns_callback) {
	struct Auth *auth = event->query->data;
	struct IODNSResult *dnsresult = event->result;
	
	auth->dnslookup = NULL;
	
	if(event->type == IODNSEVENT_SUCCESS) {
		strncpy(auth->host, dnsresult->result.host, HOSTLEN);
		socket_printf(auth->conn, "NOTICE AUTH :*** Found your hostname (%s)", auth->host);
	} else {
		struct IODNSAddress *sockaddr = iosocket_get_remote_addr(auth->conn->socket);
		iodns_print_address(sockaddr, auth->conn->socket->ipv6, auth->host, HOSTLEN);
		socket_printf(auth->conn, "NOTICE AUTH :*** Couldn't look up your hostname. Using your IP instead (%s)", auth->host);
	}
	if(dnsresult)
		iodns_free_result(dnsresult);
	
	auth_try_finish(auth);
}


void auth_try_finish(struct Auth *auth) {
	if(auth->server) {
		
	} else {
		if(!auth->have_nick || !auth->have_user)
			return;
		if(!auth->sent_ping) {
			auth->sent_ping = 1;
		} else if(auth->have_pong && auth->have_dnsresolv) {
			struct Client *client = client_connected(auth);
			auth->conn->authed = 1;
			auth->conn->data.client = client;
			auth_free(auth);
		}
	}
}

void auth_abort(struct Auth *auth) {
	
}

static void auth_free(struct Auth *auth) {
	if(auth->dnslookup) {
		
	}
	
}
